//********************************************************************************
// FILE:		main.c
// PROJECT:		zarLib  - ver 1.00 -
// COPYRIGHTS:	The famous group 4 of "Roskilde Universites Center"
//*********************************************************************************/
// Headers:
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <time.h>
#include "zarLib.h"
#include "main.h"


//********************************************************************************
// Constants:

#define NUM_COMMANDS				2
#define NUM_GENERAL_SWITCHES		2
#define NUM_COMPRESSION_SWITCHES	5

#define COMPRESS					1
#define EXTRACT						2

#define DUMP_TIME					1
#define ASSUME_YES					2


//********************************************************************************
// Gloabls:

/* Commands */
static ZAR_COMMAND g_commands[NUM_COMMANDS]					= {	{ "Compress file.", "c" },
																{ "Extract file.", "x"	}  };

/* General switches */
static ZAR_SWITCH g_genSwitches[NUM_GENERAL_SWITCHES]		= {	{ "Dump the calculation time.", "t"	  },
																{ "Assume yes on all questions.", "y" }  };

/* Compression switches */
static ZAR_SWITCH g_comSwitches[NUM_COMPRESSION_SWITCHES]	= {	{ "Use the Shannon-Fano algorithm.",						"sf" },
																{ "Use the Hoffman algorithm.\t",							"h"	 },
																{ "Use Arithmetic coding.\t\t [Not implemened]",			"ac" },
																{ "Use the LZ77 algorithm.",								"77" },
																{ "Use the LZW algorithm.\t\t",								"lw" } };

//********************************************************************************
// Function prototypes:

static int compressMain(ZAR_TASK* pzTask);
static int extract(ZAR_TASK* pzTask, ZAR_FILE* pzSrc);
static int compress(ZAR_TASK* pzTask, ZAR_FILE* pzSrc);


//********************************************************************************
// main()

int main(int nArgs, char** ppArgs)
{
	int		  nRet;
	ZAR_INIT  init;
	
	// Init library:
	init.dwFlags		= ZAR_INIT_GENERAL_SWITCHES | ZAR_INIT_COMPRESSION_SWITCHES;
	init.nNumCommands	= NUM_COMMANDS;
	init.pCommands		= g_commands;
	init.nNumGenSwitch  = NUM_GENERAL_SWITCHES;
	init.pGenSwitch		= g_genSwitches;
	init.nNumComSwitch	= NUM_COMPRESSION_SWITCHES;
	init.pComSwitch		= g_comSwitches;
	init.compressMain   = compressMain;

	zarInit(&init);

	// Run library:
	nRet = zarMain(ppArgs);

	// Close library:
	zarClose();	

	return nRet;
}


//********************************************************************************
// compressMain()

static int compressMain(ZAR_TASK* pzTask)
{
	int			nSuccess = 0;
	ZAR_FILE*	pzSrc;

	// Open the source file:
	if (NULL == (pzSrc = zarLoadInputFile(pzTask->szSrc)))
		return 0;

	if (zarFileSize(pzSrc) == 0)
		return zarError(ZAR_ERR_MESSAGE, "File is empty");
	

	// Set the update size to 1/10 of filesize:
	zarSetUpdateChunck(pzSrc->dwFileSize /10);
	
	// Extract or compress file:
	if (pzTask->dwCommand & EXTRACT)
		nSuccess = extract(pzTask, pzSrc);
	else
	if (pzTask->dwCommand & COMPRESS)
		nSuccess = compress(pzTask, pzSrc);

	if (nSuccess)
	{
		// Show used time?
		if (pzTask->dwGeneralSwitch & DUMP_TIME)
			printf("\nTime used: %0.4f seconds.", zarDumpTime() /(float)CLOCKS_PER_SEC);

		// Print space (looks better):
		printf("\n");
	}	
	
	// Flush all data to files:
	fflush(NULL);
	
	// Close input-file:
	zarCloseFile(pzSrc);

	return 1;
}


//********************************************************************************
// extract()

static int extract(ZAR_TASK* pzTask, ZAR_FILE* pzSrc)
{
	int			nRet = 0;
	_ushort		usCompressMethod;
	ZAR_FILE*	pzDst;
	
	// Get the header of the file:
	if (!zarLoadFileHeader(pzSrc, &usCompressMethod, (pzTask->szDst[0] == 0 ? pzTask->szDst : NULL)))
		return 0;

	// Create the output file:
	if (NULL == (pzDst = zarLoadOutputFile(pzTask->szDst, (pzTask->dwGeneralSwitch & ASSUME_YES ? 0 : ZAR_ASK_IF_EXISTS))))
		return 0;

	// Start time:
	zarInitTime();

	switch (usCompressMethod)
	{
		case SHANNON_FANO_COMPRESSION:
			nRet = ShannonFanoExtract(pzSrc, pzDst);
			break;

		case LZW_COMPRESSION:
			nRet = lzwExtract(pzSrc, pzDst);
			break;

		case LZ77_COMPRESSION:
			nRet = lz77Extract(pzSrc, pzDst);
			break;
		
		case HOFFMAN_COMPRESSION:
			nRet = HuffmanExtract(pzSrc, pzDst);
			break;
		
		default:
			zarError(ZAR_ERR_FEATURE_NOT_IMPLEMENTED, "compression method");
	}

	// Close the output file:
	zarCloseFile(pzDst);

	return nRet;
}


//********************************************************************************
// compress()

static int compress(ZAR_TASK* pzTask, ZAR_FILE* pzSrc)
{
	ZAR_FILE*	pzDst;
		
	// Validate the compression method:
	if (pzTask->dwCompressSwitch == 0)
		pzTask->dwCompressSwitch = DEFAULT_COMPRESSION;
	else
	if (zarGetNumSwitches(pzTask->dwCompressSwitch) != 1)
		return zarError(ZAR_ERR_MESSAGE, "Only one compression method may be specified!");	
	
	// Is their an output-file?
	if (pzTask->szDst[0] == 0)
	{
		// Then use the name of the input file:
		if (!zarGetFileNameInPath(pzTask->szSrc, pzTask->szDst))
			return zarError(ZAR_ERR_NOT_VALID_FILENAME, pzTask->szSrc);
	}

	// Add extension to output file (if nessary):
	zarSetFileExtension(pzTask->szDst);

	// Load the output-file:
	if (NULL == (pzDst = zarLoadOutputFile(pzTask->szDst, (pzTask->dwGeneralSwitch & ASSUME_YES ? 0 : ZAR_ASK_IF_EXISTS))))
		return 0;

	// Save the fileheader for the output-file:
	zarSaveFileHeader(pzDst, (_ushort) pzTask->dwCompressSwitch, pzTask->szSrc);

	// Start time:
	zarInitTime();
	
	// Launch the compression algorithms:
	switch (pzTask->dwCompressSwitch)
	{
		case SHANNON_FANO_COMPRESSION:
			if (!ShannonFanoCompress(pzSrc, pzDst))
				goto lbl_error;
			break;
	
		case LZW_COMPRESSION:
			if (!lzwCompress(pzSrc, pzDst))
				goto lbl_error;
			break;

		case LZ77_COMPRESSION:
			if (!lz77Compress(pzSrc, pzDst))
				goto lbl_error;
			break;

		case HOFFMAN_COMPRESSION:
			if (!HuffmanCompress(pzSrc, pzDst))
				goto lbl_error;
			break;
		
		default:
			zarError(ZAR_ERR_FEATURE_NOT_IMPLEMENTED, "algorithm");
			goto lbl_error;
	}

	// Show the compression rate:
	zarFileSize(pzDst);
	printf(" %02.2f%%", (1.0 -(pzDst->dwFileSize /(float)pzSrc->dwFileSize)) *100.0);
	
	// Flush all data to file:
	zarFlushFileData(pzDst);
	zarCloseFile(pzDst);
	return 1;


lbl_error:
	
	// Close file:
	zarCloseFile(pzDst);
		
	return 0;
}